/***************************************************************************
 * Project Name       : FZM
 * File name          :
 * Module name        :
 * Author             : Aswini SG
 * Date created       : 20.03.2013
 * Date modified last :
 * -------------------------------------------------------------------------
 * Description        : SE config lib
 ***************************************************************************
 * Copyright (c) Nokia 2012
 * All rights reserved.
 * Licensed material - property of Nokia
 ***************************************************************************
 * Revision History (newest ontop)
 * Date        | Name of Person | Description
 * ------------+----------------+-------------------------------------------
 **************************************************************************/

#include "iwf_table_creation.h"

extern   dl_gtp_mgmt_error_ind_hash_bucket_t * dl_gtp_mgmt_error_ind_table;
extern unsigned int iwf_log_level;
#define ERR_IND_TABLE_BIT_MASK  0x3fff
#define MAX_ERR_IND_TABLE_HASH_BUCKETS          4096

//#define __IWF_ERR_IND_DEBUG__

#define DBG_DUMP_ON 0
#if DBG_DUMP_ON
#define DBG_PRT(format, ...) printk(format,##__VA_ARGS__)
#else
#define DBG_PRT(format, ...)
#endif


uint32_t  pteid_pip_hash_get(uint32_t p_teid_dl, iwf_ip_addr_t  p_ip_dl)
{
    uint32_t key = 0;
    uint32_t hash = 0;
    uint32_t ip_dl;

    if(p_ip_dl.version == IPv6)
    {
        uint32_t ip_dl_1;
        uint32_t ip_dl_2;
        ip_dl_1 = p_ip_dl.ip6_addr32[0] + p_ip_dl.ip6_addr32[2];
        ip_dl_2 = p_ip_dl.ip6_addr32[1] + p_ip_dl.ip6_addr32[3];
        key     = (p_teid_dl + (ip_dl_1 ^ ip_dl_2));
    }
    else
    {
        ip_dl = p_ip_dl.ip4_addr32;
        key   = (p_teid_dl + ip_dl);
    }

    key  %= TEID_SICP_HASH_BUCKETS ;

    //14 bit hashing done
    hash = key & ERR_IND_TABLE_BIT_MASK;

    //printk ("Hash val is %u \n", hash);
    return hash;
}

uint32_t set_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, teid_sicp_gtpu_mgmt_data_t * teid_sicp_gtpu_mgmt_data)
{
    uint32_t hash = 0;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *head_bucket;
    unsigned int i;
    char  ip_str[40];
    int bucketIndex = 0;

    if(IWF_LOG_LEVEL_TABLES_CONFIG & iwf_log_level)
    {
        if(0 != printk_ratelimit())
        {

            printk("\nIWF_ERR_IND_DEBUG :In %s : p_ip_dl=%s, rx_nid=%u, tx_nid=%u, rx_cpid=%u, tx_cpid=%u, mtype=%u \n",
                    __FUNCTION__,
                    ((teid_sicp_gtpu_mgmt_data->p_ip.version == IPv6) ?
                     (fzm_enet_inet_ntoa_IPv6_r(teid_sicp_gtpu_mgmt_data->p_ip.addr_u.ipv6Address, ip_str)) :
                     (fzm_enet_inet_ntoa_r(teid_sicp_gtpu_mgmt_data->p_ip.addr_u.ipv4Address, ip_str))),
                    teid_sicp_gtpu_mgmt_data->sicp_data.rx_nid,
                    teid_sicp_gtpu_mgmt_data->sicp_data.tx_nid,
                    teid_sicp_gtpu_mgmt_data->sicp_data.rx_cpid,
                    teid_sicp_gtpu_mgmt_data->sicp_data.tx_cpid,
                    teid_sicp_gtpu_mgmt_data->sicp_data.mtype);
        }
    }

    if((is_ip_addr_null(&teid_sicp_gtpu_mgmt_data->p_ip) == true)    ||
            (!((teid_sicp_gtpu_mgmt_data->sicp_data.rx_nid)  &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.tx_nid)  &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.rx_cpid) &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.tx_cpid) &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.mtype)   &&
               (p_teid_dl))))
    {
        DBG_PRT("set_dl_gtp_mgmt_error_ind_connection: error values\n");
        return VTC_ERR;
    }

    hash = pteid_pip_hash_get(p_teid_dl, teid_sicp_gtpu_mgmt_data->p_ip);

    head_bucket = &dl_gtp_mgmt_error_ind_table[hash];
    bucket = head_bucket;

    DBG_PRT ("set_teid_gtpu_mgmt_hash 12: p_teid_dl = 0x%08x, p_ip_dl = 0x%x, hash = 0x%08x\n",
      p_teid_dl,
      teid_sicp_gtpu_mgmt_data->p_ip.ip4_addr32,
      hash);

    bucketIndex = hash;

    while (bucket)
    {
        for (i = 0; i < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; i++)
        {
            dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[i]);
            if ((entry->p_teid == p_teid_dl) && ip_is_equal(&entry->sicp_gtpu_mgmt_data.p_ip, &teid_sicp_gtpu_mgmt_data->p_ip))
            {
                if (entry->is_duplicate_entry == 1)
                {
                    //printk ("SE_TRACE: set_teid_gtpu_mgmt_hash: entry with same TEID/pIP already existing -> ERR\n");
                    // entry with same TEID/pIP already existing and duplicate entry flag is set.
                    
                    DBG_PRT("set_dl_gtp_mgmt_error_ind_connection: err dup values\n");
                    return ERR_DUPLICATE_TEID;
                }
                else
                {
                    // overwrite entry as duplicate entry flag is not set
                    entry->p_teid = p_teid_dl;
                    memcpy(&entry->sicp_gtpu_mgmt_data, teid_sicp_gtpu_mgmt_data, sizeof(teid_sicp_gtpu_mgmt_data_t));
                    entry->is_duplicate_entry = 0x1;
                    DBG_PRT("set_dl_gtp_mgmt_error_ind_connection: override\n");
                    return VTC_OK;
                }
            }

            if (entry->p_teid == 0)
            {
                // found free entry
                //printk("BINGO!! Found a New Entry \n");
                entry->p_teid = p_teid_dl;
                memcpy(&entry->sicp_gtpu_mgmt_data, teid_sicp_gtpu_mgmt_data, sizeof(teid_sicp_gtpu_mgmt_data_t));
                DBG_PRT("set_dl_gtp_mgmt_error_ind_connection: bucket[%d] i[%d]\n", bucketIndex, i);
                DBG_PRT("After Setting, this is - entry->p_teid =%d, IP Address is=%d\n",entry->p_teid,entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32);

                entry->is_duplicate_entry = 0x0;

                bucket->entries_gtpu_mgmt_used++;
                return VTC_OK;
            }
        }
        bucket = bucket->next;
        bucketIndex++;
    }

    bucket = head_bucket;
    while (bucket->next)
        bucket = bucket->next;

    // no free entry found in any of the linked buckets, create a new one
    bucket->next = (struct _dl_gtp_mgmt_error_ind_hash_bucket *) kmalloc (
            sizeof (struct _dl_gtp_mgmt_error_ind_hash_bucket), GFP_KERNEL);

    if (bucket->next == NULL)
    {
        //printk("Out of memory allocating memory for dl_gtp_mgmt_error_ind_hash_bucket\n");
        return VTC_ERR;
    }

    memset(bucket->next, 0, sizeof(dl_gtp_mgmt_error_ind_hash_bucket_t));

    {
        dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->next->entries_gtpu_mgmt[0]);
        entry->p_teid                                = p_teid_dl;
        memcpy(&entry->sicp_gtpu_mgmt_data, teid_sicp_gtpu_mgmt_data, sizeof(teid_sicp_gtpu_mgmt_data_t));
        bucket->next->entries_gtpu_mgmt_used++;
        DBG_PRT("set_dl_gtp_mgmt_error_ind_connection: created new bucket, and added\n");
        return VTC_OK;
    }
}  // end of set_**_gtp_mgmt_connection


uint32_t ip_se_gtp_mgmt_error_ind_table_set (sIpSicpApiMsg *msg)
{

    uint32_t                     retVal    = VTC_OK; // OK = 0 !!
    /* Down-link / ingress processing */
    ip_sicp_api_t              * cfg       = (ip_sicp_api_t *) getCfgMsgPayload(msg);

    char  ip_str[40];
    if(IWF_LOG_LEVEL_TABLES_CONFIG & iwf_log_level)
    {
        if(0 != printk_ratelimit())
        {

            printk("\n In  IP Version :%d\n",cfg->u.teid_sicp_gtpu_mgmt_data.p_ip.version);
            printk("\nIWF_ERR_IND_DEBUG :In %s : p_teid_dl=%u, p_ip_dl=%s, rx_nid=%u, tx_nid=%u, rx_cpid=%u, tx_cpid=%u, mtype=%u \n",
                    __FUNCTION__,cfg->up.p_teid_dl, 
                    ((cfg->u.teid_sicp_gtpu_mgmt_data.p_ip.version == IPv6)?
                     (fzm_enet_inet_ntoa_IPv6_r(cfg->u.teid_sicp_gtpu_mgmt_data.p_ip.addr_u.ipv6Address, ip_str)) :
                     (fzm_enet_inet_ntoa_r(cfg->u.teid_sicp_gtpu_mgmt_data.p_ip.addr_u.ipv4Address, ip_str))),
                    cfg->u.teid_sicp_gtpu_mgmt_data.sicp_data.rx_nid,
                    cfg->u.teid_sicp_gtpu_mgmt_data.sicp_data.tx_nid,
                    cfg->u.teid_sicp_gtpu_mgmt_data.sicp_data.rx_cpid,
                    cfg->u.teid_sicp_gtpu_mgmt_data.sicp_data.tx_cpid,
                    cfg->u.teid_sicp_gtpu_mgmt_data.sicp_data.mtype);
        }
    }
    retVal = set_dl_gtp_mgmt_error_ind_connection(cfg->up.p_teid_dl, (teid_sicp_gtpu_mgmt_data_t *) &cfg->u.teid_sicp_gtpu_mgmt_data);

    return retVal;
}  // end of ip_se_**_gtp_mgmt_set




teid_sicp_gtpu_mgmt_data_t * get_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, iwf_ip_addr_t p_ip_dl)
{
    uint32_t hash = 0;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    unsigned int i;

    char  ip_str[40];
    if(IWF_LOG_LEVEL_TABLES_CONFIG & iwf_log_level)
    {
        if(0 != printk_ratelimit())
        {

            printk("\nIWF_ERR_IND_DEBUG :IP received %s", p_ip_dl.version == IPv6 ? fzm_enet_inet_ntoa_IPv6_r(p_ip_dl.addr_u.ipv6Address, ip_str) : 
                    fzm_enet_inet_ntoa_r(p_ip_dl.addr_u.ipv4Address, ip_str));
        }
    }

    hash = pteid_pip_hash_get(p_teid_dl, p_ip_dl);

    //printk("SHWC p_teid_dl=%d\n",p_teid_dl);
    bucket = &dl_gtp_mgmt_error_ind_table[hash];
    //printk("Bucket=%p\n",bucket);
    while (bucket)
    {
        for (i = 0; i < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; i++)
        {
            dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[i]);
            /*printk("Found these: entry->p_teid =%d AND entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32 = %x \n",
              entry->p_teid,entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32);*/

            if ((entry->p_teid == p_teid_dl))
            {
                if (entry->sicp_gtpu_mgmt_data.p_ip.version == IPv6)
                {
                    if(ip_is_equal(&entry->sicp_gtpu_mgmt_data.p_ip, &p_ip_dl))
                    return &(entry->sicp_gtpu_mgmt_data);
                }
                else
                {
                    if(entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32 == p_ip_dl.ip4_addr32)
                        return &(entry->sicp_gtpu_mgmt_data);
                }
            }
        }
        bucket = bucket->next;
    }

    return NULL;
}  // end of get_**_gtp_mgmt_connection



uint32_t del_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, iwf_ip_addr_t  p_ip_dl)
{
    uint32_t hash = 0;
    dl_gtp_mgmt_error_ind_hash_bucket_t *head_bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    unsigned int i;

    hash = pteid_pip_hash_get(p_teid_dl, p_ip_dl);

    head_bucket = &dl_gtp_mgmt_error_ind_table[hash];
    bucket = head_bucket;

    //printk("SHWC p_teid_dl=%d\n p_ip_dl=%d\n",p_teid_dl,p_ip_dl);
    //printk("Bucket=%p\n",bucket);

    while (bucket)
    {
        //printk("In While Bucket=%p\n",bucket);
        for (i = 0; i < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; i++)
        {
            // printk("Inside FOR for %d time\n",i);
            dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[i]);
            if ((entry->p_teid == p_teid_dl) && (ip_is_equal(&entry->sicp_gtpu_mgmt_data.p_ip, &p_ip_dl))) 

            {
                //    printk("Inside IF after the data match for %d time\n",i);
                if (entry->is_duplicate_entry == 0x1)
                {
                    //  printk("Inside IF after the duplicate entry match for %d time\n",i);
                    //reset the duplicate_entry flag and return
                    entry->is_duplicate_entry = 0;

                    return VTC_OK;
                }
                else
                {
                    //printk("About to delete it for %dst time\n",i);
                    // entry found, delete it
                    memset(entry, 0 , sizeof(dl_gtp_mgmt_error_ind_hash_entry_t));
                    // printk("MEMSET DONE\n");
                    bucket->entries_gtpu_mgmt_used--;
                    if ((bucket->entries_gtpu_mgmt_used == 0) && (bucket != head_bucket))
                    {
                        // bucket is empty, if it's not the first bucket, delete it
                        // printk("BUCKET EMPTY CASE\n");
                        dl_gtp_mgmt_error_ind_hash_bucket_t *temp = head_bucket;
                        while (temp->next != bucket)
                            temp = temp->next;
                        temp->next = bucket->next;
                        memset (bucket, 0, sizeof (dl_gtp_mgmt_error_ind_hash_bucket_t));
                        // printk("MEMSET DONE\n");
                        kfree (bucket);
                        // printk("KFREE DONE\n");
                    }
                    return VTC_OK;
                }
            }
        }
        bucket = bucket->next;
    }

    return VTC_ERR;
}  // end of del_**_gtp_mgmt_connection





uint32_t ip_se_gtp_mgmt_error_ind_table_get (sIpSicpApiMsg *msg)
{
    /* Down-link / ingress processing */
    ip_sicp_api_t              * cfg       = (ip_sicp_api_t *) getCfgMsgPayload(msg);
    teid_sicp_gtpu_mgmt_data_t * conf_data = (teid_sicp_gtpu_mgmt_data_t *) &cfg->u.teid_sicp_gtpu_mgmt_data;
    uint32_t                     retVal    = VTC_OK;

    conf_data = get_dl_gtp_mgmt_error_ind_connection(cfg->up.p_teid_dl, conf_data->p_ip);

    if (conf_data != NULL)
    {
        /* Finally read values should be given back towards host, not only printed! */
        memcpy(&(cfg->u.teid_sicp_gtpu_mgmt_data), conf_data, sizeof(teid_sicp_gtpu_mgmt_data_t));

    }
    else
    {
        // printk("SE_TRACE: error -> non valid or no values configured, GTP mamagement! \n");
        retVal = VTC_ERR;
    }

    return retVal;
}  // end of ip_se_**_gtp_mgmt_get



uint32_t ip_se_gtp_mgmt_error_ind_table_del (sIpSicpApiMsg *msg)
{
    uint32_t retVal    = VTC_OK;

    /* Down-link / ingress processing */
    ip_sicp_api_t              * cfg       = (ip_sicp_api_t *) getCfgMsgPayload(msg);
    teid_sicp_gtpu_mgmt_data_t * conf_data = (teid_sicp_gtpu_mgmt_data_t *) &cfg->u.teid_sicp_gtpu_mgmt_data;

    if (conf_data == NULL)
    {
        retVal = VTC_ERR;
        return retVal;
    }

    retVal = del_dl_gtp_mgmt_error_ind_connection(cfg->up.p_teid_dl, conf_data->p_ip);

    return retVal;
}  // end of ip_se_**_gtp_mgmt_del






uint32_t mod_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, teid_sicp_gtpu_mgmt_data_t * teid_sicp_gtpu_mgmt_data)
{
    uint32_t hash = 0;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    unsigned int i;

    if((is_ip_addr_null(&teid_sicp_gtpu_mgmt_data->p_ip) == true ) ||
            (!((teid_sicp_gtpu_mgmt_data->sicp_data.rx_nid)  &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.tx_nid)  &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.rx_cpid) &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.tx_cpid) &&
               (teid_sicp_gtpu_mgmt_data->sicp_data.mtype)   &&
               (p_teid_dl))))
    {
        return VTC_ERR;
    }

    hash = pteid_pip_hash_get(p_teid_dl , teid_sicp_gtpu_mgmt_data->p_ip);

    bucket = &dl_gtp_mgmt_error_ind_table[hash];

    while (bucket)
    {
        for (i = 0; i < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; i++)
        {
            //  printk("In for mod_dl_gtp_mgmt_error_ind_connection\n");
            dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[i]);
            if ((entry->p_teid == p_teid_dl) &&
                    ip_is_equal(&entry->sicp_gtpu_mgmt_data.p_ip, &teid_sicp_gtpu_mgmt_data->p_ip))
            {
                // entry found, update it
                memcpy(&entry->sicp_gtpu_mgmt_data, teid_sicp_gtpu_mgmt_data, sizeof(teid_sicp_gtpu_mgmt_data_t));
                return VTC_OK;
            }
        }
        bucket = bucket->next;
    }

    return VTC_ERR;
}  // end of mod_**_gtp_mgmt_connection

uint32_t ip_se_gtp_mgmt_error_ind_table_mod (sIpSicpApiMsg *msg)
{
    ip_sicp_api_t              * cfg       = (ip_sicp_api_t *)               getCfgMsgPayload(msg);
    uint32_t                     retVal    = VTC_OK;


    retVal = mod_dl_gtp_mgmt_error_ind_connection(cfg->up.teid_dl, (teid_sicp_gtpu_mgmt_data_t *) &cfg->u.teid_sicp_gtpu_mgmt_data);

    return retVal;
}  // end of ip_se_**_gtp_mgmt_mod

uint32_t proc_dump_errIndTable (struct seq_file *m)
{
    dl_gtp_mgmt_error_ind_hash_bucket_t *head_bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *temp;
    unsigned int i, j, n, n4, n6;
    char ip_str[40];

    seq_printf(m,"\nLTE IWF Err Ind Table:\n");
    seq_printf(m,"===================================================================\n");
    seq_printf(m,"TEID     Peer IP                         RX_NID   TX_NID    RX_CPID    TX_CPID   MTYPE\n");
    seq_printf(m,"=======  =============================   ======   =======   =======    =======   =====\n");

    n = 0;
    n4 = 0;
    n6 = 0;
    for (i = 0; i < DL_GTP_MGMT_ERROR_IND_HASH_BUCKETS; i++)
    {
        head_bucket = &dl_gtp_mgmt_error_ind_table[i];
        bucket = head_bucket;
        temp = head_bucket;

        if(bucket)
        {
            for (j = 0; j < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; j++)
            {
                dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[j]);
                if ((entry != NULL) && entry->p_teid)
                {
                    DBG_PRT("errIndTable: bucketIndex[%d] pos[%d]\n", i, j);
                    if (entry->sicp_gtpu_mgmt_data.p_ip.version == IPv4)
                    {
                        iwf_ip_addr_t ip;
                        ip.addr_u.ipv4Address.ipa_32 = htonl(entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32);
                        seq_printf(m,"0x%08x  %-15s  0x%04x  0x%04x  0x%04x  0x%04x  0x%04x\n",
                                entry->p_teid,
                                fzm_enet_inet_ntoa_r(ip.addr_u.ipv4Address, ip_str),
                                entry->sicp_gtpu_mgmt_data.sicp_data.rx_nid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.tx_nid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.rx_cpid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.tx_cpid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.mtype);
                        n4++;

                    }
                    else if (entry->sicp_gtpu_mgmt_data.p_ip.version == IPv6)
                    {
                        iwf_ip_addr_t ip;
                        ip.addr_u.ipv6Address.ipa_32[0] = htonl(entry->sicp_gtpu_mgmt_data.p_ip.addr_u.ipv6Address.ipa_32[0]);
                        ip.addr_u.ipv6Address.ipa_32[1] = htonl(entry->sicp_gtpu_mgmt_data.p_ip.addr_u.ipv6Address.ipa_32[1]);
                        ip.addr_u.ipv6Address.ipa_32[2] = htonl(entry->sicp_gtpu_mgmt_data.p_ip.addr_u.ipv6Address.ipa_32[2]);
                        ip.addr_u.ipv6Address.ipa_32[3] = htonl(entry->sicp_gtpu_mgmt_data.p_ip.addr_u.ipv6Address.ipa_32[3]);

                        seq_printf(m,"0x%08x    %-39s  0x%04x  0x%04x  0x%04x   0x%04x   0x%04x\n",
                                entry->p_teid,
                                fzm_enet_inet_ntoa_IPv6_r(ip.addr_u.ipv6Address,ip_str),
                                entry->sicp_gtpu_mgmt_data.sicp_data.rx_nid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.tx_nid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.rx_cpid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.tx_cpid,
                                entry->sicp_gtpu_mgmt_data.sicp_data.mtype);
                        n6++;

                    }
                    else
                    {
                        seq_printf(m,"IP version not set !!!\n");                         

                    }							
                    n++;
                }
            }
        }
        else
        {
            continue;
        }
        //bucket = temp->next;
    }

    seq_printf(m,"Total in err ind table: %u : IPv4[%u] IPv6[%u]\n", n, n4, n6);                         

    return VTC_OK;
}

